const expect = require('chai').expect;
const sinon = require('sinon');

const varsnap = require('../varsnap/core');
const ConfigTools = require('./config_tools');

function example(x, y) {
  return x * y;
}

describe('Util', () => {
  describe('getConfig', () => {
    beforeEach(ConfigTools.saveConfigs);
    afterEach(ConfigTools.restoreConfigs);
    it('can read from env vars', () => {
      ConfigTools.writeConfig('env', 'test');
      const config = varsnap.Util.getConfig('env');
      expect(config).to.equal('test');
    });
    it('will lowercase configs', () => {
      ConfigTools.writeConfig('env', 'TEST');
      const config = varsnap.Util.getConfig('env');
      expect(config).to.equal('test');
    });
  });

  describe('getSignature', () => {
    it('returns a valid signature', () => {
      const signature = varsnap.Util.getSignature(varsnap.Util.getConfig);
      expect(signature).to.include('javascript');
      expect(signature).to.include(varsnap.version);
      expect(signature).to.include('getConfig');
    });
  });
});

describe('Producer', () => {
  beforeEach(ConfigTools.saveConfigs);
  afterEach(ConfigTools.restoreConfigs);
  describe('isEnabled', () => {
    beforeEach(() => {
      ConfigTools.writeConfig(varsnap.configKeys.varsnap, 'true');
      ConfigTools.writeConfig(varsnap.configKeys.env, 'production');
      ConfigTools.writeConfig(varsnap.configKeys.producerToken, 'asdf');
    });
    it('will return false if varsnap is not enabled', () => {
      ConfigTools.writeConfig(varsnap.configKeys.varsnap, 'false');
      expect(varsnap.Producer.isEnabled()).to.be.false;
    });
    it('will return false if env is not production', () => {
      ConfigTools.writeConfig(varsnap.configKeys.env, 'development');
      expect(varsnap.Producer.isEnabled()).to.be.false;
    });
    it('will return false if producer token is not set', () => {
      ConfigTools.writeConfig(varsnap.configKeys.producerToken, '');
      expect(varsnap.Producer.isEnabled()).to.be.false;
    });
    it('will return true if all configs are correct', () => {
      expect(varsnap.Producer.isEnabled()).to.be.true;
    });
  });
  describe('serialize', () => {
    it('can serialize data', () => {
      expect(varsnap.Producer.serialize(5)).to.equal('5');
      expect(varsnap.Producer.serialize({'1':[2]})).to.equal('{"1":[2]}');
    });
  });
  describe('produce', () => {
    beforeEach(() => {
      ConfigTools.writeConfig(varsnap.configKeys.varsnap, 'true');
      ConfigTools.writeConfig(varsnap.configKeys.env, 'production');
      ConfigTools.writeConfig(varsnap.configKeys.producerToken, 'asdf');
      this.ajax = sinon.stub(varsnap.Util, 'ajax');
    });
    afterEach(() => {
      this.ajax.restore();
    });
    it('will send data', () => {
      this.ajax.resolves('asdf');
      const producer = new varsnap.Producer(example);
      producer.produce([2, 3], 6);
      sinon.assert.called(this.ajax);
    });
  });
});

describe('Consumer', () => {
  beforeEach(ConfigTools.saveConfigs);
  afterEach(ConfigTools.restoreConfigs);
  describe('isEnabled', () => {
    beforeEach(() => {
      ConfigTools.writeConfig(varsnap.configKeys.varsnap, 'true');
      ConfigTools.writeConfig(varsnap.configKeys.env, 'development');
      ConfigTools.writeConfig(varsnap.configKeys.consumerToken, 'asdf');
    });
    it('will return false if varsnap is not enabled', () => {
      ConfigTools.writeConfig(varsnap.configKeys.varsnap, 'false');
      expect(varsnap.Consumer.isEnabled()).to.be.false;
    });
    it('will return false if env is not development', () => {
      ConfigTools.writeConfig(varsnap.configKeys.env, 'production');
      expect(varsnap.Consumer.isEnabled()).to.be.false;
    });
    it('will return false if consumer token is not set', () => {
      ConfigTools.writeConfig(varsnap.configKeys.consumerToken, '');
      expect(varsnap.Consumer.isEnabled()).to.be.false;
    });
    it('will return true if all configs are correct', () => {
      expect(varsnap.Consumer.isEnabled()).to.be.true;
    });
  });
  describe('deserialize', () => {
    it('can deserialize data', () => {
      expect(varsnap.Consumer.deserialize('5')).to.equal(5);
      expect(varsnap.Consumer.deserialize('{"1":[2]}')).to.deep.equal({'1':[2]});
    });
  });
  describe('consume', () => {
    beforeEach(() => {
      ConfigTools.writeConfig(varsnap.configKeys.varsnap, 'true');
      ConfigTools.writeConfig(varsnap.configKeys.env, 'development');
      ConfigTools.writeConfig(varsnap.configKeys.consumerToken, 'asdf');
      this.clock = sinon.useFakeTimers();
      this.ajax = sinon.stub(varsnap.Util, 'ajax');
      this.consumer = new varsnap.Consumer(example);
    });
    afterEach(() => {
      this.clock.restore();
      this.ajax.restore();
    });
    it('will not run if the consumer is not enabled', () => {
      ConfigTools.writeConfig(varsnap.configKeys.env, 'production');
      this.consumer.consume();
      this.clock.tick(1001);
      sinon.assert.callCount(this.ajax, 0);
    });
    it('will make an ajax request and report it', () => {
      this.consumer.consume();
      const data = {
        'id': 'abcd',
        'inputs': '[2, 3]',
        'prod_outputs': '4',
      };
      this.ajax.resolves(data);
      this.clock.tick(1001);
      sinon.assert.called(this.ajax);
    });
  });
});

describe('varsnap', () => {
  it('creates a consumer and producer', () => {
    const wrapped = varsnap.varsnap(example);
    expect(wrapped.producer.targetFunc).to.equal(example);
    expect(wrapped.consumer.targetFunc).to.equal(example);
  });
  it('wraps a function', () => {
    const wrapped = varsnap.varsnap(example);
    expect(wrapped(2, 4)).to.equal(8);
  });
});

describe('version', () => {
  it('has a version', () => {
    expect(varsnap.version).to.be;
  });
});

