const request = require('request-promise-native');
const deepEqual = require('fast-deep-equal');

const producerURL = 'https://www.varsnap.com/api/snap/produce/';
const consumerURL = 'https://www.varsnap.com/api/snap/consume/';
const version = 'v0.1.1';

const configKeys = {
  varsnap: 'varsnap',
  env: 'env',
  producerToken: 'varsnapProducerToken',
  consumerToken: 'varsnapConsumerToken',
};

class Util {
  static getConfig(config) {
    let val = '';
    if (typeof window === 'undefined') {
      // node.js
      val = process && process.env && process.env[config];
    } else {
      // browser
      val = window.varsnap && window.varsnap[config];
    }
    if (!val) {
      return '';
    }
    val = val.toLowerCase();
    return val;
  }

  static getSignature(targetFunc) {
    let signature = targetFunc.toString().split('{')[0];
    signature = signature.replace(/^\s+|\s+$/g, '');
    return 'javascript.' + version + '.' + signature;
  }

  static ajax(url, requestData) {
    const options = {
      method: 'POST',
      uri: url,
      form: requestData,
      json: true,
      followAllRedirects: true,
    };
    const ajaxCall = request(options);
    return ajaxCall;
  }
}

class Producer {
  constructor(targetFunc) {
    this.targetFunc = targetFunc;
  }

  static isEnabled() {
    if (Util.getConfig(configKeys.varsnap) !== 'true') {
      return false;
    }
    if (Util.getConfig(configKeys.env) !== 'production') {
      return false;
    }
    if (!Util.getConfig(configKeys.producerToken)) {
      return false;
    }
    return true;
  }

  static serialize(data) {
    return JSON.stringify(data);
  }

  produce(args, output) {
    if (!Producer.isEnabled()) {
      return;
    }
    console.log('Sending call to Varsnap'); // eslint-disable-line no-console
    const data = {
      producer_token: Util.getConfig(configKeys.producerToken),
      signature: Util.getSignature(this.targetFunc),
      inputs: Producer.serialize(args),
      prod_outputs: Producer.serialize(output),
    };
    Util.ajax(producerURL, data).catch((err) => {
      if(err.statusCode === 404) {
        console.error('Cannot produce to Varsnap; check your producer token'); // eslint-disable-line no-console
      } else {
        console.error(err.error); // eslint-disable-line no-console
      }
    });
  }
}

class Consumer {
  constructor(targetFunc) {
    this.targetFunc = targetFunc;
    this.lastSnapId = '';
  }

  static isEnabled() {
    if (Util.getConfig(configKeys.varsnap) !== 'true') {
      return false;
    }
    if (Util.getConfig(configKeys.env) !== 'development') {
      return false;
    }
    if (!Util.getConfig(configKeys.consumerToken)) {
      return false;
    }
    return true;
  }

  static deserialize(data) {
    return JSON.parse(data);
  }

  consume() {
    if (!Consumer.isEnabled()) {
      return;
    }
    console.log('Streaming in production Varsnap function calls...'); // eslint-disable-line no-console
    setInterval(this.consumeOnce.bind(this), 1000);
  }

  consumeOnce() {
    const data = {
      consumer_token: Util.getConfig(configKeys.consumerToken),
      signature: Util.getSignature(this.targetFunc),
    };
    Util.ajax(consumerURL, data).then((data) => {
      if(!data || data['id'] == this.lastSnapId) {
        return;
      }
      this.lastSnapId = data['id'];
      console.log('Receiving call from Varsnap uuid: ' + this.lastSnapId); // eslint-disable-line no-console
      const inputs = Consumer.deserialize(data['inputs']);
      const prodOutputs = Consumer.deserialize(data['prod_outputs']);
      const localOutputs = this.targetFunc(...inputs);
      this.report(inputs, prodOutputs, localOutputs);
    }).catch((error) => {
      console.error('Cannot contact Varsnap' + error); // eslint-disable-line no-console
    });
  }

  report(inputs, prodOutputs, localOutputs) {
    const functionName = Util.getSignature(this.targetFunc);
    console.log('Function:                      ' + functionName);   // eslint-disable-line no-console
    console.log('Function input args:           ' + inputs);         // eslint-disable-line no-console
    console.log('Production function outputs:   ' + prodOutputs);    // eslint-disable-line no-console
    console.log('Your function outputs:         ' + localOutputs);   // eslint-disable-line no-console
    const matches = deepEqual(prodOutputs, localOutputs);
    console.log('Matching outputs:              ' + matches);        // eslint-disable-line no-console
    console.log('');                                                 // eslint-disable-line no-console
  }
}

function varsnap(func) {
  const producer = new Producer(func);
  const consumer = new Consumer(func);

  consumer.consume();

  const wrapped = function(...args) {
    let result = '';
    try {
      result = func(...args);
    }
    catch(err) {
      producer.produce(args, err);
      throw(err);
    }
    producer.produce(args, result);
    return result;
  };

  wrapped.producer = producer;
  wrapped.consumer = consumer;
  return wrapped;
}

module.exports = {
  configKeys: configKeys,
  Util: Util,
  Producer: Producer,
  Consumer: Consumer,
  varsnap: varsnap,
  version: version,
};
